; ========================================================================
; Bootsector for 1.44MB floppies
; ------------------------------------------------------------------------
; Adapted from:
;	RetroForth/Native 8 Revision D, by Charles Childers
; 	Original F Bootsector, by Robert Oestling
; ------------------------------------------------------------------------
; Changelog:
; 03/18/05
;	Chris Walton
;	* Stopped using multitrack reads, for older BIOSes (from F)
; 	* Changed to inline print, for better code readability
; 03/16/05
;	Chris Walton
;	* Changed GDT to be Intel compliant
; 	* Smaller nicer A20 code
;	* Now works as a drop-in replacement for RF8/Native!
; 03/13/05
;	Chris Walton
;	* Got it started. Displays welcome message.
; ------------------------------------------------------------------------
; NOTICE: This is free software. Use it for whatever you want, modify it 
; to suit your needs, but I ask you to contribute your changes back to the
; community for the better good. There's no guarantee, so if this software
; damages you or your property in anyway you cannot hold me or anybody 
; who modified it for a non-nefarious purpose responsible.
; ========================================================================

; Tell assembler what we're doing here..
use16
org	0x7c00
format	binary

; Start of bootsector
start:	
	; Some BIOSes have a bug where CS is not zero when starting up
	; the bootsector. This jong jump will make sure.
	jmp	0x0000:.start

	; Zero all other segment registers, too.
.start:	xor	ax, ax
	mov	ds, ax
	mov	es, ax
	mov	ss, ax

	; Set up stack
	mov	sp, 0x7000

	; Display welcome message
	call	print16
	db	"Loading...", 0

	; Each disk operation has 3 retries
	mov	si, 3

	; We're loading to 0x7e00, so we set the segment
	; to 0x07e0 and leave bx zero (easier that way).
	mov	ax, 0x07e0
	mov	es, ax
	xor	bx, bx

 	; Drive 0, head 0, track 0, sector 2, count 17
	; (that means, every sector except bootsector)
	mov	cx, 2
	xor	dx, dx	
.retry1:mov	ax, ((2 shl 8) + 30)

	; Do it, jump to error handler on error, otherwise "return".
	int	0x13
	jnc	.done1

	; Decrease error count and retry if not zero
	dec	si
	jnz	.retry1
	
	; Display error message and hang
.ffail:	call	print16
	db	"Failed due to floppy error!", 0
	call	motoroff
	jmp	$

	; Now, read the second track in its entirety
	; 17 times 512 bytes read
.done1:	add	bx, (512*30)
	mov	cx, 1
	mov	si, 3
.retry2:mov	ax, ((2 shl 8) + 31)
	
	; Read it, on success, jump ahead, else do error thing
	int	0x13
	jnc	.done2
	dec	si
	jnz	.retry2

	; Message and hang
	jmp	.ffail

.done2:	call	motoroff
	
	; Switch to protected mode
	; Clear interrupts
	cli

	; Enable A20
	call	a20enable

	; Load the GDT, and set the pmode bit
	lgdt	[gdt]
	mov	eax, cr0
	inc	eax
	mov	cr0, eax

	; Jump to pmode code
	jmp	dword 8:pmode	

; Enable A20 gate
; Taken from F by Robert Oestling
a20enable:
	mov	al,0xd1
	out	0x64,al
	call	a20wait
	mov	al,0xdf
	out	0x60,al
	call	a20wait
	mov	al,0xff
	out	0x64,al
a20wait:
	in	al,0x64
	jmp	.l0
.l0:	and	al,2
	jnz	a20wait
	ret

; Write a string
; The string is zero terminated and located at return address.
; Doesn't save registers (why should it?)
print16:
	; Grab address of string	
	pop	si

	; Grab byte, check for zero
.again:	lodsb
	or	al, al
	jz	.done

	; Print the char, and loop
	mov	ah, 0x0e
	mov	bx, 14
	int	0x10
	jmp	.again

	; String offset is now return address
.done:	jmp	si

	
; Turn floppy motor off, the hard way.
motoroff:
	xor	al, al
	mov	dx, 0x3f2
	out	dx, al
	ret

; The previous GDT worked, but didn't conform to
; how it was supposed to work. So, I took the GDT
; from F and fixed it up to work here. It is still
; a mystery to me why the old GDT worked.
gdt:
	dw	(.end-.dummy)
	dd	.dummy
	dw	0
.dummy:	dw	0,0,0,0
	dw	0xffff,0x0000
	db	0x00,0x9a,0xcf,0x00
	dw	0xffff,0x0000
	db	0x00,0x92,0xcf,0x00
.end:

; PROTECTED MODE CODE STARTS HERE!
use32

; Set up and get to kernel
pmode:
	; Init segments (again) and stack
	mov	ax, 0x10
	mov	ds, ax
	mov	es, ax
	mov	ss, ax
	mov	esp, 0x50000
	
	; Move kernel from 0x7e00 to 0x110000
	; (stupid real mode stuff, argh, I hate it)
	cld
	mov	esi, 0x7e00
	mov	edi, 0x11000
	mov	edx, edi
	mov	ecx, (_end-kernel)/4
	rep	movsd

	; Call the kernel, and hang if theres a problem
	call	edx
	jmp	$

; Pad the rest of the bootsector, and add the magic bytes
; (Absolutely unnecessary crap, some systems don't even 
; check for it, but .... *sigh* )
times   510-($-0x7c00)	    db 0
dw 0xAA55

; Include precompiled image AFTER the 512 bytes
kernel:
	align 4
	file "native.bin"
	align 4
_end:
